import analyzer.DocAnalyzer;
import structures.KNN;
import structures.NaiveBayesian;
import structures.Post;
import utils.Utils;

import java.io.*;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;

/***
 * Feature selection and classification
 */
public class MP3Main {

    public static void main(String[] args) throws IOException {
        String tokenFile = "./data/Model/en-token.bin";
        String stopwords = "./data/Model/initStopWords.txt";
        String dataFile = "./data/yelp/";

        int Ngram = 1;
        int classNumber = 2;
        int lengthThreshold = 5; // >5

        int hashNum = 5;

        DocAnalyzer analyzer = new DocAnalyzer(tokenFile, Ngram, classNumber, lengthThreshold, hashNum);
        //task 1
//        featureSelectSample(analyzer, dataFile, stopwords);
        //task 2
        NBSample(analyzer, dataFile);
        //task 3
//        KNNSample(analyzer, dataFile);

    }

    /*** task 1: feature selection using CHI or IG ***/
    public static void featureSelectSample(DocAnalyzer analyzer, String dataFile, String stopwords){
        int DFThreshold = 10; //>= 10
        String featureSelection = "IG"; // CHI or IG
        double threshold = 3.841; // >=3.841 for CHI
        if (featureSelection.equals("IG")) {
            threshold = 0;// > 0 for IG
        }
        String featureFile = "./data/out/feature_" + featureSelection + ".txt";

        analyzer.LoadStopwords(stopwords);
        analyzer.LoadDirectory(dataFile, ".json", "TRAIN");
        System.out.println("corpus size: " + analyzer.getReviews().size());

        /*** Task 1: Feature Selection ***/
        System.out.println("Selecting features ...");
        HashMap<String, Double> rawFeature = analyzer.selectFeature(featureSelection, DFThreshold, -1);
        Utils.saveBoth2File(featureFile, Utils.cutProb(Utils.sortByValue(new HashMap<Object, Object>(rawFeature)),
                "VALUE", threshold));
    }

    /*** task 2: Naive Bayesian ***/
    public static void NBSample(DocAnalyzer analyzer, String dataFile){
        String cv = "./data/out/vocab.txt"; // built from feature selection task
        String proNB = "./data/out/NaiveBayesPara.txt";
        //collect statistics and calculate TFIDF
        analyzer.LoadCV(cv);
        analyzer.LoadDirectory(String.format("%strain/", dataFile), ".json", "TRAIN");
        analyzer.LoadDirectory(String.format("%stest/", dataFile), ".json", "TEST");
        ArrayList<Post> trainset = new ArrayList<>();
        ArrayList<Post> testset = new ArrayList<>();
        for(Post p : analyzer.getReviews()){
            if(p.getType().equals("TRAIN"))
                trainset.add(p);
            else
                testset.add(p);
        }

        for(int i = 0; i < 5; i++) {
            double delta = 0.1 + 0.2 * i;
            System.out.format("************* delta: %f *********\n", delta);
            NaiveBayesian nb = new NaiveBayesian(1, 0.1, delta, analyzer.getTokenStats(),
                    analyzer.getClassStats(), analyzer.getCV().size());
            // train
            Utils.saveProbList2File(proNB, Utils.sortByValue(new HashMap<Object, Object>(nb.trainNB(analyzer.getTokenStats())) ));//train

            //test
            String auc = String.format("./data/out/auc_%.1f_.txt", delta);
            plotAUC(nb.testNB(testset), auc);
        }
    }

    /*** task 3: KNN with or without random projection ***/
    public static void KNNSample(DocAnalyzer analyzer, String dataFile)throws IOException{
        String cv = "./data/out/vocab.txt"; // built from feature selection task
        String mode = "BRUTE"; //or APPR

        //collect statistics and calculate TFIDF
        analyzer.LoadCV(cv);
        analyzer.LoadDirectory(String.format("%strain/", dataFile), ".json", "TRAIN");
        analyzer.calcTFIDF();
        System.out.println("TFIDF for KNN finished");
        analyzer.getReviews().clear();
        //represent documents in TFIDF
        analyzer.representDocuments(String.format("%strain/", dataFile), ".json", "TRAIN");
        analyzer.representDocuments(String.format("%stest/", dataFile), ".json", "TEST");

        ArrayList<Post> trainset = new ArrayList<>();
        ArrayList<Post> testset = new ArrayList<>();
        for(Post p : analyzer.getReviews()){
            if(p.getType().equals("TRAIN"))
                trainset.add(p);
            else
                testset.add(p);
        }

        String folder = String.format("./data/out/KNN_%s/",mode);

        for(int l = 5; l < 10; l += 5){
            for (int k = 5; k < 10; k += 5){
                System.out.println("************* l: " + l + "; k: " + k);
                KNN knn = new KNN(l, k, analyzer.getCV());
                if (mode.equals("APPR"))
                    knn.calcProjection(analyzer.getReviews());

                NumberFormat formatter = new DecimalFormat("#0.00000");
                long start = System.currentTimeMillis();
                knn.printKNN(folder, mode, trainset, testset);
                long end1 = System.currentTimeMillis();
                System.out.println(mode + "KNN cost: "
                        + formatter.format((end1 - start) / 1000d) + " seconds");

            }
        }
    }

    public static double[][] crossValidation(DocAnalyzer analyzer, int foldNum, String model){

        int foldSize = analyzer.getReviews().size() / foldNum;
        double[][] results = new double[foldNum+1][3];
        double sum1=0, sum2=0, sum3=0;

        Collections.shuffle(analyzer.getReviews());
        for(int k = 0;k < foldNum; k++){
            System.out.println("***** fold: " + k );
            List<Post> testSet = new ArrayList<Post>(analyzer.getReviews().subList(k * foldSize, Math.min((k+1) * foldSize,
                    analyzer.getReviews().size())));
            System.out.println("test size: " + testSet.size());
            List<Post> trainSet = new ArrayList<Post>(analyzer.getReviews().subList(0, Math.min(k * foldSize ,
                    analyzer.getReviews().size())));
            trainSet.addAll(new ArrayList<Post>(analyzer.getReviews().subList(Math.min((k+1) * foldSize, analyzer.getReviews().size()),
                    analyzer.getReviews().size())));
            System.out.println("train size: " + trainSet.size());

            List<Integer> predLabel = new ArrayList<>();
            if(model.equals("KNN-APPR")){
                KNN knn = new KNN(5, 5, analyzer.getCV());
                predLabel = knn.approKNN(trainSet, testSet);
            }else if(model.equals("NB")){
                NaiveBayesian nb = new NaiveBayesian(1, 0.1, 0.1,
                        analyzer.getTokenStats(), analyzer.getClassStats(), analyzer.getCV().size());
                predLabel = nb.classifyNB(trainSet, testSet);
            }

            int TP = 0;
            int FN = 0;
            int FP = 0;
            for(int i = 0; i < predLabel.size(); i++){
                if(predLabel.get(i) == 1 && testSet.get(i).getRating()>= 4){
                    TP++;
                }else if(predLabel.get(i) == 1 && testSet.get(i).getRating() < 4){
                    FP++;
                }else if(predLabel.get(i) == 0 && testSet.get(i).getRating() >= 4){
                    FN++;
                }
            }
            double recall = (double)TP/(TP+FN);
            double precision = (double)TP/(TP+FP);
            double F1 = 2/(1/recall + 1/precision);
            results[k][0] = recall;
            results[k][1] = precision;
            results[k][2] = F1;
            sum1 += recall;
            sum2 += precision;
            sum3 += F1;
        }
        sum1 = sum1 / foldNum;
        sum2 = sum2 / foldNum;
        sum3 = sum3 / foldNum;
        results[foldNum][0] = sum1;
        results[foldNum][1] = sum2;
        results[foldNum][2] = sum3;

        System.out.println("***** Results for " + model);
        System.out.println("k-fold\trecall\tprecision\tF1");
        for(int k = 0; k < foldNum; k++){
            System.out.println(String.format("%d\t%.5f\t%.5f\t%.5f", k, results[k][0], results[k][1], results[k][2]));
        }
        System.out.println(String.format("%s\t%.5f\t%.5f\t%.5f", "ave", sum1, sum2, sum3));
        return results;
    }

    public static void plotAUC(HashMap<Post, Double> score, String outfile){
        List<Map.Entry<Object, Object>> scoreRank = Utils.sortByValue(new HashMap<Object, Object>(score));
        HashMap<Double, Double> AUC = new HashMap<>();
        for(int i = 0; i < scoreRank.size(); i++){
            int TP = 0;
            int FN = 0;
            int FP = 0;
            // larger then 1
            List<Map.Entry<Object, Object>> largerHalf = scoreRank.subList(0, i+1);
            for(int j = 0; j < largerHalf.size(); j++){
                int predLabel = 1;
                int trueLabel = ((Post)largerHalf.get(j).getKey()).getRating()<4? 0:1;
                if(trueLabel == 1)
                    TP++;
                else
                    FP++;
            }
            // smaller than 0
            List<Map.Entry<Object, Object>> smallerHalf = scoreRank.subList(i+1, scoreRank.size());
            for(int j = 0; j < smallerHalf.size(); j++){
                int predLabel = 0;
                int trueLabel = ((Post) smallerHalf.get(j).getKey()).getRating()<4? 0:1;
                if(trueLabel == 1)
                    FN++;
            }

            AUC.put((double)TP/(TP+FN), (double) TP/(TP+FP));
        }
        Utils.saveBoth2File(outfile, Utils.sortByValue(new HashMap<Object, Object>(AUC)));
    }

}
