package structures;

import structures.Token;
import utils.Utils;

import java.io.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class FeatureSelector {
    int m_minDF = 10, m_maxDF;
    HashMap<String, Double> m_selectedFeatures;

    public FeatureSelector(int minDF, int maxDF){
        m_maxDF = maxDF<=0 ? Integer.MAX_VALUE : maxDF;
        m_minDF = minDF;
        m_selectedFeatures = new HashMap<>();
    }

    public HashMap<String, Double> getSelectedFeatures(){
        return m_selectedFeatures;
    }

    // information gain
    public void IG(HashMap<String, Token> featureStat, int[] classStat){
        m_selectedFeatures.clear();
        double DFSumC = Utils.sumOfArray(classStat);
        double[] proC = new double[classStat.length];
        double[] proC_T = new double[classStat.length];
        double[] proC_noT = new double[classStat.length];

        double entropySum = 0;
        for(int c = 0; c < classStat.length; c++){
            proC[c] = classStat[c] / DFSumC;
            if(proC[c] != 0){
                entropySum -= proC[c] * Math.log(proC[c]);
            }
        }

        for(String t : featureStat.keySet()){
            double entropySum_t = 0, entropySum_not = 0;
            double IGt = 0;

            Token ftStat = featureStat.get(t);
            double DFSumT = Utils.sumOfArray(ftStat.getValues());
            double proT = 0, proNoT = 0;
            if(DFSumT >= m_minDF && DFSumT < m_maxDF){
                proT = DFSumT / DFSumC;
                proNoT = 1 - proT;

                for(int i = 0; i < classStat.length; i++){
                    double proT_C = (double) ftStat.getValues()[i] / classStat[i];
                    proC_T[i] = proT_C * proC[i] / proT;
                    proC_noT[i] = (1-proT_C) * proC[i] / proNoT;
                    if(proC_T[i] != 0){
                        entropySum_t += proC_T[i] * Math.log(proC_T[i]);
                    }
                    if(proC_noT[i] != 0){
                        entropySum_not += proC_noT[i] * Math.log(proC_noT[i]);
                    }
                }

                IGt = entropySum + proT * entropySum_t + proNoT * entropySum_not;
                m_selectedFeatures.put(t, IGt);
            }
        }
    }

    // Chi Square
    public void CHI(HashMap<String, Token> featureStat, int[] classStat){
        m_selectedFeatures.clear();

        double sumAll = Utils.sumOfArray(classStat);
        for(String t : featureStat.keySet()){
            Token ftStat = featureStat.get(t);
            double DFSumT = Utils.sumOfArray(ftStat.getValues());
            if(DFSumT >= m_minDF && DFSumT < m_maxDF){
                double chiAve = 0;
                for(int i = 0; i < classStat.length; i++){
                    double proC = classStat[i] / sumAll;
                    double A = ftStat.getValues()[i]; // c & t
                    double B = classStat[i] - A; // c & !t
                    double C = DFSumT - A; // !c & t
                    double D = sumAll - A - B - C;// !c & !t
                    double chi = sumAll * (A*D - B*C) * (A*D - B*C)
                            / ((A+C) * (B+D) * (A+B) * (C+D));
                    chiAve += chi * proC;
                }
                m_selectedFeatures.put(t, chiAve);
            }
        }
    }
}
