package structures;

import java.io.*;
import java.util.*;
import utils.*;

public class KNN {
    int m_k;
    private int m_bitNum;
    private HashMap<String, Integer> m_CV;
    double[][] m_base;

    public KNN(int l, int k, HashMap<String, Integer> CV){
        this.m_k = k;
        this.m_bitNum = l;
        this.m_CV = CV;
        this.m_base = new double[l][m_CV.size()];
        //generate base
        Random rd = new Random();
        for(int i = 0; i < l; i++){
            for (int j = 0; j < m_CV.size(); j++){
                this.m_base[i][j] = rd.nextDouble() * 2 - 1;
            }
        }
    }

    // generate random projection
    public void randomProjection(Post review){
        int[] hash = new int[m_bitNum];
        HashMap<String, Token> vct = review.getVct();
        for(int i = 0; i < m_bitNum; i++){
            double dotProduct = 0;
            for(String word : vct.keySet()){
                dotProduct += vct.get(word).getValues()[0]
                        * m_base[i][m_CV.get(word)];
            }
            if(dotProduct < 0)
                hash[i] = 0;
            else
                hash[i] = 1;
        }
        review.setHash(hash);
    }

    // APPR: approximate KNN
    // BRUTE: exact KNN
    HashMap<Post, Double> runKNN(List<Post> trainSet, Post test, String mode){
        HashMap<Post, Double> distance = new HashMap<>();
        for(Post p : trainSet){
            if(mode.equals("APPR")) {
                distance.put(p, p.similiarityHash(test));
            }else if(mode.equals("BRUTE")){
                distance.put(p, p.similiarity(test));
            }
        }
        return distance;
    }

    public void calcProjection(ArrayList<Post> reviews){
        for(Post p : reviews){
            randomProjection(p); //update review with m_hashCode
        }
    }

    public List<Integer> approKNN(List<Post> trainSet, List<Post> testSet){
        List<Integer> labels = new ArrayList<>();
        for(Post review : testSet){
            HashMap<Post, Double> dis = runKNN(trainSet, review, "APPR");
            List<Map.Entry<Object, Object>> neighbors = Utils.cutProb(Utils.sortByValue(new HashMap<Object, Object>(dis)),
                    "COUNT", m_k);
            int k0 = 0;
            int k = neighbors.size();
            for(Map.Entry<Object, Object> ngb : neighbors){
                Post p = (Post) ngb.getKey();
                if(p.getRating() < 4){
                    k0++;
                }
            }
            if(k0 > k-k0){
                labels.add(0);
            }else{
                labels.add(1);
            }
        }
        return labels;
    }

    public void printKNN(String folder, String mode, List<Post> trainSet, List<Post> testSet){
        new File(folder).mkdirs();
        for(int i = 0; i < testSet.size(); i++){
            Post curDoc = testSet.get(i);
            HashMap<Post, Double> dis = runKNN(trainSet, curDoc, mode);
            String file = folder + curDoc.getID() + "_knn.txt";
            Utils.saveBoth2File(file, Utils.cutProb(Utils.sortByValue(new HashMap<Object, Object>(dis)),
                    "COUNT", m_k));
        }
    }
}
