package structures;

import utils.Utils;

import java.io.*;
import java.util.*;

public class NaiveBayesian {

    LanguageModel m_LM_pos;// to store p(w|y=1)
    LanguageModel m_LM_neg;// to store p(w|y=0)

    // controlled vocabulary
    int m_V;

    public NaiveBayesian(int Ngram, double lambda, double delta,
                         HashMap<String, Token> tokenStats, int[] classStats, int cvsize){
        m_V = cvsize;
        m_LM_pos = new LanguageModel(Ngram, lambda, delta);
        m_LM_neg = new LanguageModel(Ngram, lambda, delta);
        m_LM_pos.setModel(tokenStats, classStats, 1, m_V);
        m_LM_neg.setModel(tokenStats, classStats, 0, m_V);
    }


    public HashMap<String, Double> trainNB(HashMap<String, Token> trainSet){
        HashMap<String, Double> tokenProb = new HashMap<String, Double>();
        for(String token : trainSet.keySet()){
            tokenProb.put(token, Math.log(m_LM_pos.calcAdditiveSmoothedProb(token)
                    / m_LM_neg.calcAdditiveSmoothedProb(token)));
        }
        return tokenProb;
    }

    public HashMap<Post, Double> testNB(ArrayList<Post> testSet){
        HashMap<Post, Double> postScore = new HashMap<Post, Double>();
        for(Post p : testSet){
            double fX = Math.log((double)m_LM_pos.getDocSize() / m_LM_neg.getDocSize());
            for(String token : p.getVct().keySet()){
                fX += Utils.sumOfArray(p.getVct().get(token).getValues())
                        * Math.log(m_LM_pos.calcAdditiveSmoothedProb(token)
                        / m_LM_neg.calcAdditiveSmoothedProb(token));
            }
            postScore.put(p, fX);
        }
        return postScore;
    }

    public List<Integer> classifyNB(List<Post> trainSet, List<Post> testSet){
        List<Integer> labels = new ArrayList<>();
        for(Post p : testSet){
            double fX = Math.log((double)m_LM_pos.getDocSize() / m_LM_neg.getDocSize());
            for(String token : p.getVct().keySet()){
                fX += Utils.sumOfArray(p.getVct().get(token).getValues())
                        * Math.log(m_LM_pos.calcAdditiveSmoothedProb(token)
                        / m_LM_neg.calcAdditiveSmoothedProb(token));
            }
            if(fX < 0){
                labels.add(0);
            }else{
                labels.add(1);
            }
        }
        return labels;
    }
}
